// 
// Author:
//       smdn <smdn@mail.invisiblefulmoon.net>
// 
// Copyright (c) 2008-2010 smdn
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections.Generic;
using System.IO;
using System.Security.Cryptography;
using System.Text;

using Smdn.Collections;
using Smdn.Formats;
using Smdn.IO;
using Smdn.Net.Imap4.Client.Session;

namespace Smdn.Net.Imap4.Client {
  public class ImapMessageInfo : ImapMessageInfoBase, IImapMessageAttribute {
    public long Sequence {
      get; private set;
    }

    public long Uid {
      get; private set;
    }

    public Uri Url {
      get { return uriBuilder.Uri; }
    }

    /*
     * IImapMessageDynamicAttribute
     */
    private IImapMessageDynamicAttribute dynamicAttribute;

    internal IImapMessageDynamicAttribute DynamicAttribute {
      get { UpdateDynamicAttributes(false); return dynamicAttribute; }
      set { dynamicAttribute = value; }
    }

    public IImapMessageFlagSet Flags {
      get { return DynamicAttribute.Flags; }
    }

    [CLSCompliant(false)]
    public ulong ModSeq {
      get { return DynamicAttribute.ModSeq; }
    }

    /*
     * IImapMessageStaticAttribute
     */
    private IImapMessageStaticAttribute staticAttribute;

    internal IImapMessageStaticAttribute StaticAttribute {
      get { UpdateStaticAttributes(); return staticAttribute; }
      set { staticAttribute = value; }
    }

    public IImapBodyStructure BodyStructure {
      get { return StaticAttribute.BodyStructure; }
    }

    public ImapEnvelope Envelope {
      get { return StaticAttribute.Envelope; }
    }

    public DateTimeOffset? InternalDate {
      get { return StaticAttribute.InternalDate; }
    }

    public long Length {
      get { return StaticAttribute.Rfc822Size; }
    }

    long IImapMessageStaticAttribute.Rfc822Size {
      get { return StaticAttribute.Rfc822Size; }
    }

    long IImapMessageStaticAttribute.GetBinarySizeOf(string section)
    {
      return StaticAttribute.GetBinarySizeOf(section);
    }

    /*
     * miscellaneous getter properties
     */
    public bool IsMarkedAsDeleted {
      get { return Flags.Has(ImapMessageFlag.Deleted); }
    }

    public bool IsSeen {
      get { return Flags.Has(ImapMessageFlag.Seen); }
    }

    public bool IsRecent {
      get { return Flags.Has(ImapMessageFlag.Recent); }
    }

    private Lazy<string> envelopeSubject;

    public string EnvelopeSubject {
      get { return envelopeSubject.Value; }
    }

    private Lazy<DateTimeOffset?> envelopeDate;

    public DateTimeOffset? EnvelopeDate {
      get { return envelopeDate.Value; }
    }

    private ImapSequenceSet uidSet;

    private ImapSequenceSet SequenceOrUidSet {
      get
      {
        if (uidSet == null)
          uidSet = ImapSequenceSet.CreateUidSet(Uid);

        return uidSet;
      }
    }

    internal ImapMessageInfo(ImapOpenedMailboxInfo mailbox, long uid, long sequence)
      : base(mailbox)
    {
      this.Uid = uid;
      this.Sequence = sequence;

      (this as IImapUrl).SetBaseUrl(mailbox.Mailbox.UrlBuilder);

      InitializeLazy();
    }

    private ImapUriBuilder uriBuilder;

    protected override ImapSequenceSet GetSequenceOrUidSet()
    {
      return SequenceOrUidSet;
    }

    private void InitializeLazy()
    {
      envelopeSubject = new Lazy<string>(delegate {
        try {
          return TextConvert.FromMimeEncodedStringNullable(StaticAttribute.Envelope.Subject);
        }
        catch (FormatException) {
          return StaticAttribute.Envelope.Subject;
        }
      });

      envelopeDate = new Lazy<DateTimeOffset?>(delegate {
        try {
          return DateTimeConvert.FromRFC822DateTimeOffsetStringNullable(StaticAttribute.Envelope.Date);
        }
        catch (FormatException) {
          return default(DateTimeOffset);
        }
      });
    }

    /*
     * operations
     */
    public void Refresh()
    {
      UpdateStaticAttributes();
      UpdateDynamicAttributes(true);
    }

    internal static readonly ImapFetchDataItem FetchStaticDataItem =
      ImapFetchDataItem.BodyStructure +
      ImapFetchDataItem.Envelope + 
      ImapFetchDataItem.InternalDate +
      ImapFetchDataItem.Rfc822Size;

    private void UpdateStaticAttributes()
    {
      if (staticAttribute != null)
        return;

      Mailbox.CheckSelected();
      Mailbox.CheckUidValidity(UidValidity, SequenceOrUidSet);

      ImapMessageStaticAttribute[] staticAttr;

      ImapClient.ThrowIfError(Client.Session.Fetch(SequenceOrUidSet,
                                                   FetchStaticDataItem,
                                                   out staticAttr));

      if (0 < staticAttr.Length)
        staticAttribute = staticAttr[0];
    }

    internal static ImapFetchDataItem GetFetchDynamicDataItem(ImapCapabilityList serverCapabilities)
    {
      if (serverCapabilities.Has(ImapCapability.CondStore))
        return ImapFetchDataItem.Flags + ImapFetchDataItem.ModSeq;
      else
        return ImapFetchDataItem.Flags;
    }

    private void UpdateDynamicAttributes(bool refresh)
    {
      if (!refresh && dynamicAttribute != null)
        return;

      Mailbox.CheckSelected();
      Mailbox.CheckUidValidity(UidValidity, SequenceOrUidSet);

      ImapMessageDynamicAttribute[] dynamicAttr;

      ImapClient.ThrowIfError(Client.Session.Fetch(SequenceOrUidSet,
                                                   GetFetchDynamicDataItem(Client.ServerCapabilities),
                                                   out dynamicAttr));

      if (0 < dynamicAttr.Length)
        dynamicAttribute = dynamicAttr[0];
    }

    public void ToggleFlags(ImapMessageFlag flag, params ImapMessageFlag[] flags)
    {
      ToggleFlags(new ImapMessageFlagList(flags.Prepend(flag)));
    }

    public void ToggleKeywords(string keyword, params string[] keywords)
    {
      ToggleFlags(new ImapMessageFlagList(keywords.Prepend(keyword)));
    }

    public void ToggleFlags(IImapMessageFlagSet flagsAndKeywords)
    {
      var addFlags = new ImapMessageFlagList();
      var removeFlags = new ImapMessageFlagList();

      foreach (var flag in flagsAndKeywords) {
        if (dynamicAttribute.Flags.Has(flag))
          removeFlags.Add(flag);
        else
          addFlags.Add(flag);
      }

      if (0 < removeFlags.Count)
        Store(ImapStoreDataItem.RemoveFlags(removeFlags));

      if (0 < addFlags.Count)
        Store(ImapStoreDataItem.AddFlags(addFlags));
    }

    /*
    public override void Store(ImapStoreDataItem storeDataItem)
    {
      // TODO: update dynamic attributes
      base.Store(storeDataItem);
    }
    */

    /*
     * utility methods for Open*(), Read*()
     */
    private static string GetSection(IImapBodyStructure structure)
    {
      if (structure == null)
        throw new ArgumentNullException("structure");

      return structure.Section;
    }

    private static readonly Encoding iso8859_1 = Encoding.GetEncoding("ISO-8859-1");

    private static string GetEncoding(IImapBodyStructure structure, ImapMessageFetchBodyOptions options)
    {
      Encoding discard;

      return GetEncoding(structure, options, out discard);
    }

    private static string GetEncoding(IImapBodyStructure structure, ImapMessageFetchBodyOptions options, out Encoding charset)
    {
      var s = structure as ImapSinglePartBodyStructure;

      if (s != null && (int)(options & ImapMessageFetchBodyOptions.DecodeContent) != 0) {
        string charsetString;

        if (s.Parameters.TryGetValue("charset", out charsetString))
          charset = EncodingUtils.GetEncodingThrowException(charsetString);
        else
          charset = iso8859_1;

        return s.Encoding;
      }
      else {
        charset = null;

        return null;
      }
    }

    /*
     * OpenRead()
     */
    public Stream OpenRead(IImapBodyStructure section)
    {
      return OpenRead(GetSection(section), null, null, ImapMessageFetchBodyOptions.Default);
    }

    public Stream OpenRead(IImapBodyStructure section, ImapMessageFetchBodyOptions options)
    {
      return OpenRead(GetSection(section), null, GetEncoding(section, options), options);
    }

    public Stream OpenRead()
    {
      return OpenRead(string.Empty, null, null, ImapMessageFetchBodyOptions.Default);
    }

    public Stream OpenRead(ImapMessageFetchBodyOptions options)
    {
      return OpenRead(string.Empty, null, GetEncoding(BodyStructure, options), options);
    }

    public Stream OpenRead(string section)
    {
      return OpenRead(section, null, null, ImapMessageFetchBodyOptions.Default);
    }

    public Stream OpenRead(string section, ImapPartialRange? range)
    {
      return OpenRead(section, range, null, ImapMessageFetchBodyOptions.Default);
    }

    public Stream OpenRead(string section, ImapMessageFetchBodyOptions options)
    {
      return OpenRead(section, null, null, options);
    }

    public Stream OpenRead(string section, ImapPartialRange? range, ImapMessageFetchBodyOptions options)
    {
      return OpenRead(section, range, null, options);
    }

    private Stream OpenRead(string section, ImapPartialRange? range, string encoding, ImapMessageFetchBodyOptions options)
    {
      Mailbox.CheckSelected();
      Mailbox.CheckUidValidity(UidValidity, SequenceOrUidSet);

      var stream = new ImapFetchMessageBodyStream(Client.Session,
                                                  (int)(options & ImapMessageFetchBodyOptions.Peek) != 0,
                                                  SequenceOrUidSet,
                                                  section,
                                                  range,
                                                  ImapFetchMessageBodyStream.DefaultFetchBlockSize);

      IImapMessageAttribute discard;

      ImapClient.ThrowIfError(stream.Prepare(null, out discard));

      if (encoding != null)
        return GetDecodedStream(stream, encoding);
      else
        return stream;
    }

    private Stream GetDecodedStream(Stream stream, string encoding)
    {
      // TODO: use Smdn.Formats.Mime.ContentStreamUtils
      switch (encoding.ToLowerInvariant()) {
        case "7bit":
        case "8bit":
        case "binary":
          return stream;

        case "base64":
          return new CryptoStream(stream, new FromBase64Transform(FromBase64TransformMode.IgnoreWhiteSpaces), CryptoStreamMode.Read);

        case "quoted-printable":
          return new CryptoStream(stream, new FromQuotedPrintableTransform(), CryptoStreamMode.Read);

#if false
        case "uuencode":
        case "x-uu":
        case "x-uue":
        case "x-uuencode":
        case "x-uuencoded":
        case "gzip64":
        case "x-gzip64":
#endif
        default:
          throw new NotSupportedException(string.Format("unsupported content transfer encoding: '{0}'", encoding));
      }
    }

    /*
     * ReadAs<Stream>()
     */
    public TOutput ReadAs<TOutput>(IImapBodyStructure section, Converter<Stream, TOutput> converter)
    {
      return ReadAsCore(OpenRead(section), converter);
    }

    public TOutput ReadAs<TOutput>(IImapBodyStructure section, ImapMessageFetchBodyOptions options, Converter<Stream, TOutput> converter)
    {
      return ReadAsCore(OpenRead(section, options), converter);
    }

    public TOutput ReadAs<TOutput>(Converter<Stream, TOutput> converter)
    {
      return ReadAsCore(OpenRead(), converter);
    }

    public TOutput ReadAs<TOutput>(ImapMessageFetchBodyOptions options, Converter<Stream, TOutput> converter)
    {
      return ReadAsCore(OpenRead(options), converter);
    }

    public TOutput ReadAs<TOutput>(string section, Converter<Stream, TOutput> converter)
    {
      return ReadAsCore(OpenRead(section), converter);
    }

    public TOutput ReadAs<TOutput>(string section, ImapMessageFetchBodyOptions options, Converter<Stream, TOutput> converter)
    {
      return ReadAsCore(OpenRead(section, options), converter);
    }

    public TOutput ReadAs<TOutput>(string section, ImapPartialRange? range, Converter<Stream, TOutput> converter)
    {
      return ReadAsCore(OpenRead(section, range), converter);
    }

    public TOutput ReadAs<TOutput>(string section, ImapPartialRange? range, ImapMessageFetchBodyOptions options, Converter<Stream, TOutput> converter)
    {
      return ReadAsCore(OpenRead(section, range, options), converter);
    }

    private TOutput ReadAsCore<TOutput>(Stream stream, Converter<Stream, TOutput> converter)
    {
      using (stream) {
        if (converter == null)
          throw new ArgumentNullException("converter");

        return converter(stream);
      }
    }

    /*
     * ReadAllBytes()
     */
    public byte[] ReadAllBytes(IImapBodyStructure section)
    {
      return ReadAs<byte[]>(section, ReadAllBytesProc);
    }

    public byte[] ReadAllBytes(IImapBodyStructure section, ImapMessageFetchBodyOptions options)
    {
      return ReadAs<byte[]>(section, options, ReadAllBytesProc);
    }

    public byte[] ReadAllBytes()
    {
      return ReadAs<byte[]>(ReadAllBytesProc);
    }

    public byte[] ReadAllBytes(ImapMessageFetchBodyOptions options)
    {
      return ReadAs<byte[]>(options, ReadAllBytesProc);
    }

    public byte[] ReadAllBytes(string section)
    {
      return ReadAs<byte[]>(section, ReadAllBytesProc);
    }

    public byte[] ReadAllBytes(string section, ImapMessageFetchBodyOptions options)
    {
      return ReadAs<byte[]>(section, options, ReadAllBytesProc);
    }

    public byte[] ReadAllBytes(string section, ImapPartialRange? range)
    {
      return ReadAs<byte[]>(section, range, ReadAllBytesProc);
    }

    public byte[] ReadAllBytes(string section, ImapPartialRange? range, ImapMessageFetchBodyOptions options)
    {
      return ReadAs<byte[]>(section, range, options, ReadAllBytesProc);
    }

    private static byte[] ReadAllBytesProc(Stream stream)
    {
      return stream.ReadToEnd();
    }

    /*
     * Save()
     */
    public void Save(string path, IImapBodyStructure section)
    {
      SaveCore(path, delegate { return OpenRead(section); });
    }

    public void Save(string path, IImapBodyStructure section, ImapMessageFetchBodyOptions options)
    {
      SaveCore(path, delegate { return OpenRead(section, options); });
    }

    public void Save(string path)
    {
      SaveCore(path, delegate { return OpenRead(); });
    }

    public void Save(string path, ImapMessageFetchBodyOptions options)
    {
      SaveCore(path, delegate { return OpenRead(options); });
    }

    public void Save(string path, string section)
    {
      SaveCore(path, delegate { return OpenRead(section); });
    }

    public void Save(string path, string section, ImapMessageFetchBodyOptions options)
    {
      SaveCore(path, delegate { return OpenRead(section, options); });
    }

    public void Save(string path, string section, ImapPartialRange? range)
    {
      SaveCore(path, delegate { return OpenRead(section, range); });
    }

    public void Save(string path, string section, ImapPartialRange? range, ImapMessageFetchBodyOptions options)
    {
      SaveCore(path, delegate { return OpenRead(section, range, options); });
    }

    private void SaveCore(string path, Func<Stream> openRead)
    {
      using (var fileStream = File.OpenWrite(path)) {
        using (var stream = openRead()) {
          stream.WriteToEnd(fileStream, ImapFetchMessageBodyStream.DefaultFetchBlockSize);
        }
      }
    }

    /*
     * OpenText
     */
    public StreamReader OpenText(IImapBodyStructure section)
    {
      return OpenText(section, ImapMessageFetchBodyOptions.Default);
    }

    public StreamReader OpenText(IImapBodyStructure section, ImapMessageFetchBodyOptions options)
    {
      Encoding charset;
      var stream = OpenRead(GetSection(section), null, GetEncoding(section, options, out charset), options);

      return new StreamReader(stream, charset ?? iso8859_1);
    }

    public StreamReader OpenText()
    {
      return OpenText(string.Empty, iso8859_1, ImapMessageFetchBodyOptions.Default);
    }

    public StreamReader OpenText(ImapMessageFetchBodyOptions options)
    {
      return OpenText(string.Empty, iso8859_1, options);
    }

    public StreamReader OpenText(string section)
    {
      return OpenText(section, iso8859_1, ImapMessageFetchBodyOptions.Default);
    }

    public StreamReader OpenText(string section, ImapMessageFetchBodyOptions options)
    {
      return OpenText(section, iso8859_1, options);
    }

    public StreamReader OpenText(Encoding charset)
    {
      return OpenText(string.Empty, charset, ImapMessageFetchBodyOptions.Default);
    }

    public StreamReader OpenText(Encoding charset, ImapMessageFetchBodyOptions options)
    {
      return OpenText(string.Empty, charset, options);
    }

    public StreamReader OpenText(string section, Encoding charset)
    {
      return OpenText(section, charset, ImapMessageFetchBodyOptions.Default);
    }

    public StreamReader OpenText(string section, Encoding charset, ImapMessageFetchBodyOptions options)
    {
      if (charset == null)
        throw new ArgumentNullException("charset");

      return new StreamReader(OpenRead(section, null, null, options), charset);
    }

    /*
     * ReadAs<StreamReader>()
     */
    public TOutput ReadAs<TOutput>(Converter<StreamReader, TOutput> converter)
    {
      return ReadAsCore(OpenText(), converter);
    }

    public TOutput ReadAs<TOutput>(ImapMessageFetchBodyOptions options, Converter<StreamReader, TOutput> converter)
    {
      return ReadAsCore(OpenText(options), converter);
    }

    public TOutput ReadAs<TOutput>(IImapBodyStructure section, Converter<StreamReader, TOutput> converter)
    {
      return ReadAsCore(OpenText(section), converter);
    }

    public TOutput ReadAs<TOutput>(IImapBodyStructure section, ImapMessageFetchBodyOptions options, Converter<StreamReader, TOutput> converter)
    {
      return ReadAsCore(OpenText(section, options), converter);
    }

    public TOutput ReadAs<TOutput>(string section, Converter<StreamReader, TOutput> converter)
    {
      return ReadAsCore(OpenText(section), converter);
    }

    public TOutput ReadAs<TOutput>(string section, ImapMessageFetchBodyOptions options, Converter<StreamReader, TOutput> converter)
    {
      return ReadAsCore(OpenText(section, options), converter);
    }

    public TOutput ReadAs<TOutput>(Encoding charset, Converter<StreamReader, TOutput> converter)
    {
      return ReadAsCore(OpenText(charset), converter);
    }

    public TOutput ReadAs<TOutput>(Encoding charset, ImapMessageFetchBodyOptions options, Converter<StreamReader, TOutput> converter)
    {
      return ReadAsCore(OpenText(charset, options), converter);
    }

    public TOutput ReadAs<TOutput>(string section, Encoding charset, Converter<StreamReader, TOutput> converter)
    {
      return ReadAsCore(OpenText(section, charset), converter);
    }

    public TOutput ReadAs<TOutput>(string section, Encoding charset, ImapMessageFetchBodyOptions options, Converter<StreamReader, TOutput> converter)
    {
      return ReadAsCore(OpenText(section, charset, options), converter);
    }

    private TOutput ReadAsCore<TOutput>(StreamReader reader, Converter<StreamReader, TOutput> converter)
    {
      try {
        if (converter == null)
          throw new ArgumentNullException("converter");

        return converter(reader);
      }
      finally {
        reader.Close();
      }
    }

    /*
     * ReadLines()
     */
    public IEnumerable<string> ReadLines(IImapBodyStructure structure)
    {
      return ReadAs<IEnumerable<string>>(structure, ReadLinesProc);
    }

    public IEnumerable<string> ReadLines(IImapBodyStructure structure, ImapMessageFetchBodyOptions options)
    {
      return ReadAs<IEnumerable<string>>(structure, options, ReadLinesProc);
    }

    public IEnumerable<string> ReadLines()
    {
      return ReadAs<IEnumerable<string>>(ReadLinesProc);
    }

    public IEnumerable<string> ReadLines(ImapMessageFetchBodyOptions options)
    {
      return ReadAs<IEnumerable<string>>(options, ReadLinesProc);
    }

    public IEnumerable<string> ReadLines(string section)
    {
      return ReadAs<IEnumerable<string>>(section, ReadLinesProc);
    }

    public IEnumerable<string> ReadLines(string section, ImapMessageFetchBodyOptions options)
    {
      return ReadAs<IEnumerable<string>>(section, options, ReadLinesProc);
    }

    public IEnumerable<string> ReadLines(Encoding charset)
    {
      return ReadAs<IEnumerable<string>>(charset, ReadLinesProc);
    }

    public IEnumerable<string> ReadLines(Encoding charset, ImapMessageFetchBodyOptions options)
    {
      return ReadAs<IEnumerable<string>>(charset, options, ReadLinesProc);
    }

    public IEnumerable<string> ReadLines(string section, Encoding charset)
    {
      return ReadAs<IEnumerable<string>>(section, charset, ReadLinesProc);
    }

    public IEnumerable<string> ReadLines(string section, Encoding charset, ImapMessageFetchBodyOptions options)
    {
      return ReadAs<IEnumerable<string>>(section, charset, options, ReadLinesProc);
    }

    private static IEnumerable<string> ReadLinesProc(StreamReader reader)
    {
      for (;;) {
        var line = reader.ReadLine();

        if (line == null)
          yield break;
        else
          yield return line;
      }
    }

    /*
     * ReadAllLines()
     */
    public string[] ReadAllLines(IImapBodyStructure structure)
    {
      return ReadLines(structure).ToArray();
    }

    public string[] ReadAllLines(IImapBodyStructure structure, ImapMessageFetchBodyOptions options)
    {
      return ReadLines(structure, options).ToArray();
    }

    public string[] ReadAllLines()
    {
      return ReadLines().ToArray();
    }

    public string[] ReadAllLines(ImapMessageFetchBodyOptions options)
    {
      return ReadLines(options).ToArray();
    }

    public string[] ReadAllLines(string section)
    {
      return ReadLines(section).ToArray();
    }

    public string[] ReadAllLines(string section, ImapMessageFetchBodyOptions options)
    {
      return ReadLines(section, options).ToArray();
    }

    public string[] ReadAllLines(Encoding charset)
    {
      return ReadLines(charset).ToArray();
    }

    public string[] ReadAllLines(Encoding charset, ImapMessageFetchBodyOptions options)
    {
      return ReadLines(charset, options).ToArray();
    }

    public string[] ReadAllLines(string section, Encoding charset)
    {
      return ReadLines(section, charset).ToArray();
    }

    public string[] ReadAllLines(string section, Encoding charset, ImapMessageFetchBodyOptions options)
    {
      return ReadLines(section, charset, options).ToArray();
    }

    /*
     * ReadAllText()
     */
    public string ReadAllText(IImapBodyStructure structure)
    {
      return ReadAs<string>(structure, ReadAllTextProc);
    }

    public string ReadAllText(IImapBodyStructure structure, ImapMessageFetchBodyOptions options)
    {
      return ReadAs<string>(structure, options, ReadAllTextProc);
    }

    public string ReadAllText()
    {
      return ReadAs<string>(ReadAllTextProc);
    }

    public string ReadAllText(ImapMessageFetchBodyOptions options)
    {
      return ReadAs<string>(options, ReadAllTextProc);
    }

    public string ReadAllText(string section)
    {
      return ReadAs<string>(section, ReadAllTextProc);
    }

    public string ReadAllText(string section, ImapMessageFetchBodyOptions options)
    {
      return ReadAs<string>(section, options, ReadAllTextProc);
    }

    public string ReadAllText(Encoding charset)
    {
      return ReadAs<string>(charset, ReadAllTextProc);
    }

    public string ReadAllText(Encoding charset, ImapMessageFetchBodyOptions options)
    {
      return ReadAs<string>(charset, options, ReadAllTextProc);
    }

    public string ReadAllText(string section, Encoding charset)
    {
      return ReadAs<string>(section, charset, ReadAllTextProc);
    }

    public string ReadAllText(string section, Encoding charset, ImapMessageFetchBodyOptions options)
    {
      return ReadAs<string>(section, charset, options, ReadAllTextProc);
    }

    private static string ReadAllTextProc(StreamReader reader)
    {
      return reader.ReadToEnd();
    }

    /// <summary>Infrastructure. It is not intended to be used directly from your code.</summary>
    void IImapUrl.SetBaseUrl(ImapUriBuilder baseUrl)
    {
      uriBuilder = baseUrl.Clone();

      uriBuilder.Uid = Uid;
      uriBuilder.UidValidity = UidValidity;
    }

    public override string ToString()
    {
      return string.Format("{{ImapMessageInfo: Authority='{0}', Mailbox='{1}', Sequence={2}, Uid={3}, UidValidity={4}}}",
                           ImapStyleUriParser.GetStrongAuthority(Url),
                           Mailbox.FullName,
                           Sequence,
                           Uid,
                           UidValidity);
    }
  }
}
