using System;
using NUnit.Framework;
using NUnit.Framework.SyntaxHelpers;

namespace Smdn.Net.Imap4.Client {
  [TestFixture]
  public class ImapMessageInfoBaseTests {
    private class MessageSet : ImapMessageInfoBase {
      public MessageSet(ImapOpenedMailboxInfo mailbox, ImapSequenceSet sequenceOrUidSet)
        : base(mailbox)
      {
        this.sequenceOrUidSet = sequenceOrUidSet;
      }

      protected override ImapSequenceSet GetSequenceOrUidSet()
      {
        return sequenceOrUidSet;
      }

      private ImapSequenceSet sequenceOrUidSet;
    }

    private void TestMessage(ImapSequenceSet sequenceOrUidSet, Action<ImapPseudoServer, ImapMessageInfoBase> action)
    {
      TestMessage(null, sequenceOrUidSet, action);
    }

    private void TestMessage(ImapCapability[] capabilities, ImapSequenceSet sequenceOrUidSet, Action<ImapPseudoServer, ImapMessageInfoBase> action)
    {
      TestUtils.TestAuthenticated(capabilities, delegate(ImapPseudoServer server, ImapClient client) {
        // LIST
        server.EnqueueTaggedResponse("* LIST () \".\" INBOX\r\n" +
                                     "$tag OK done\r\n");
        // SELECT
        server.EnqueueTaggedResponse("* OK [UIDVALIDITY 1]\r\n" +
                                     "$tag OK done\r\n");

        using (var inbox = client.OpenInbox()) {
          server.DequeueRequest(); // LIST
          server.DequeueRequest(); // SELECT

          try {
            var messages = new MessageSet(inbox, sequenceOrUidSet);

            Assert.AreSame(messages.Client, client);
            Assert.AreSame(messages.Mailbox, inbox);
            Assert.AreEqual(1L, messages.UidValidity);

            action(server, messages);
          }
          finally {
            if (inbox.IsOpen)
              // CLOSE
              server.EnqueueTaggedResponse("$tag OK done\r\n");
          }
        }
      });
    }

    private ImapSequenceSet CreateEmptySet()
    {
      return ImapSequenceSet.CreateSet(new long[0]);
    }

    [Test]
    public void TestAddFlags()
    {
      TestMessage(ImapSequenceSet.CreateUidSet(1L), delegate(ImapPseudoServer server, ImapMessageInfoBase message) {
        // UID STORE
        server.EnqueueTaggedResponse("$tag OK done\r\n");

        message.AddFlags(ImapMessageFlag.Draft);
        // TODO: fetch updated attributes

        Assert.That(server.DequeueRequest(), Text.EndsWith("UID STORE 1 +FLAGS (\\Draft)\r\n"));
      });
    }

    [Test]
    public void TestRemoveFlags()
    {
      TestMessage(ImapSequenceSet.CreateUidSet(1L),
                  delegate(ImapPseudoServer server, ImapMessageInfoBase message) {
        // UID STORE
        server.EnqueueTaggedResponse("$tag OK done\r\n");

        message.RemoveFlags(ImapMessageFlag.Draft);
        // TODO: fetch updated attributes

        Assert.That(server.DequeueRequest(), Text.EndsWith("UID STORE 1 -FLAGS (\\Draft)\r\n"));
      });
    }

    [Test]
    public void TestReplaceFlags()
    {
      TestMessage(ImapSequenceSet.CreateUidSet(1L),
                  delegate(ImapPseudoServer server, ImapMessageInfoBase message) {
        // UID STORE
        server.EnqueueTaggedResponse("$tag OK done\r\n");

        message.ReplaceFlags(ImapMessageFlag.Draft, ImapMessageFlag.Flagged);
        // TODO: fetch updated attributes

        Assert.That(server.DequeueRequest(), Text.EndsWith("UID STORE 1 FLAGS (\\Draft \\Flagged)\r\n"));
      });
    }

    [Test]
    public void TestStoreEmptySet()
    {
      TestMessage(CreateEmptySet(),
                  delegate(ImapPseudoServer server, ImapMessageInfoBase message) {
        // do nothing
        message.Store(ImapStoreDataItem.AddFlags(ImapMessageFlag.Seen));
      });
    }

    [Test]
    public void TestStoreUidSingle()
    {
      TestMessage(ImapSequenceSet.CreateUidSet(1L),
                  delegate(ImapPseudoServer server, ImapMessageInfoBase message) {
        // UID STORE
        server.EnqueueTaggedResponse("$tag OK done\r\n");

        message.Store(ImapStoreDataItem.AddFlags(ImapMessageFlag.Seen));
        // TODO: fetch updated attributes

        Assert.That(server.DequeueRequest(), Text.EndsWith("UID STORE 1 +FLAGS (\\Seen)\r\n"));
      });
    }

    [Test]
    public void TestStoreSequenceRangeSet()
    {
      TestMessage(ImapSequenceSet.CreateRangeSet(1L, 10L),
                  delegate(ImapPseudoServer server, ImapMessageInfoBase messages) {
        // STORE
        server.EnqueueTaggedResponse("$tag OK done\r\n");

        messages.Store(ImapStoreDataItem.RemoveFlags(ImapMessageFlag.Seen));
        // TODO: fetch updated attributes

        Assert.That(server.DequeueRequest(), Text.EndsWith("STORE 1:10 -FLAGS (\\Seen)\r\n"));
      });
    }

    [Test, ExpectedException(typeof(ImapMailboxClosedException))]
    public void TestStoreMailboxClosed()
    {
      TestMessage(ImapSequenceSet.CreateRangeSet(1L, 10L),
                  delegate(ImapPseudoServer server, ImapMessageInfoBase messages) {
        // CLOSE
        server.EnqueueTaggedResponse("$tag OK done\r\n");

        messages.Mailbox.Close();

        Assert.IsFalse(messages.Mailbox.IsOpen);

        messages.Store(ImapStoreDataItem.RemoveFlags(ImapMessageFlag.Seen));
      });
    }

    [Test, Ignore("not implemented")]
    public void TestStoreUidValidityChanged()
    {
    }

    [Test, Ignore("not implemented")]
    public void TestStoreNonApplicableFlags()
    {
    }

    [Test]
    public void TestMarkAsDeleted()
    {
      TestMessage(ImapSequenceSet.CreateRangeSet(1L, 10L),
                  delegate(ImapPseudoServer server, ImapMessageInfoBase messages) {
        // STORE
        server.EnqueueTaggedResponse("$tag OK done\r\n");

        messages.MarkAsDeleted();
        // TODO: fetch updated attributes

        Assert.That(server.DequeueRequest(), Text.EndsWith("STORE 1:10 +FLAGS (\\Deleted)\r\n"));
      });
    }

    [Test]
    public void TestMarkAsSeen()
    {
      TestMessage(ImapSequenceSet.CreateRangeSet(1L, 10L),
                  delegate(ImapPseudoServer server, ImapMessageInfoBase messages) {
        // STORE
        server.EnqueueTaggedResponse("$tag OK done\r\n");

        messages.MarkAsSeen();
        // TODO: fetch updated attributes

        Assert.That(server.DequeueRequest(), Text.EndsWith("STORE 1:10 +FLAGS (\\Seen)\r\n"));
      });
    }

    [Test]
    public void TestDeleteUidPlusCapable()
    {
      Delete(true);
    }

    [Test]
    public void TestDeleteUidPlusIncapable()
    {
      Delete(false);
    }

    private void Delete(bool uidPlusCapable)
    {
      var capabilities = uidPlusCapable
        ? new[] {ImapCapability.UidPlus}
        : new ImapCapability[0];

      TestMessage(capabilities,
                  ImapSequenceSet.CreateUidRangeSet(1L, 10L),
                  delegate(ImapPseudoServer server, ImapMessageInfoBase messages) {
        // STORE
        server.EnqueueTaggedResponse("$tag OK done\r\n");
        // EXPUNGE/UID EXPUNGE
        server.EnqueueTaggedResponse("$tag OK done\r\n");

        messages.Delete();
        // TODO: fetch updated attributes

        Assert.That(server.DequeueRequest(), Text.EndsWith("STORE 1:10 +FLAGS (\\Deleted)\r\n"));
        if (uidPlusCapable)
          Assert.That(server.DequeueRequest(), Text.EndsWith("UID EXPUNGE 1:10\r\n"));
        else
          Assert.That(server.DequeueRequest(), Text.EndsWith("EXPUNGE\r\n"));
      });
    }

    [Test]
    public void TestDeleteEmptySet()
    {
      TestMessage(CreateEmptySet(),
                  delegate(ImapPseudoServer server, ImapMessageInfoBase messages) {
        // do nothing
        messages.Delete();
      });
    }

    [Test]
    public void TestCopyToDestinationMailbox()
    {
      CopyOrMoveToDestinationMailbox(false);
    }

    [Test]
    public void TestMoveToDestinationMailbox()
    {
      CopyOrMoveToDestinationMailbox(true);
    }

    private void CopyOrMoveToDestinationMailbox(bool move)
    {
      TestMessage(ImapSequenceSet.CreateUidSet(1L, 3L, 5L),
                  delegate(ImapPseudoServer server, ImapMessageInfoBase messages) {
        // LIST
        server.EnqueueTaggedResponse("* LIST () \"\" dest\r\n" +
                                     "$tag OK done\r\n");

        var destMailbox = messages.Client.GetMailbox("dest");

        server.DequeueRequest();

        // UID COPY
        server.EnqueueTaggedResponse("$tag OK done\r\n");
        if (move)
          // UID STORE
          server.EnqueueTaggedResponse("$tag OK done\r\n");

        if (move)
          messages.MoveTo(destMailbox);
        else
          messages.CopyTo(destMailbox);
        // TODO: fetch updated attributes

        Assert.That(server.DequeueRequest(), Text.EndsWith("UID COPY 1,3,5 \"dest\"\r\n"));

        if (move)
          Assert.That(server.DequeueRequest(), Text.EndsWith("UID STORE 1,3,5 +FLAGS (\\Deleted)\r\n"));
      });
    }

    [Test]
    public void TestMoveToDestinationMailboxCopyFailed()
    {
      TestMessage(ImapSequenceSet.CreateUidSet(1L, 3L, 5L),
                  delegate(ImapPseudoServer server, ImapMessageInfoBase messages) {
        // LIST
        server.EnqueueTaggedResponse("* LIST () \"\" dest\r\n" +
                                     "$tag OK done\r\n");

        var destMailbox = messages.Client.GetMailbox("dest");

        server.DequeueRequest();

        // UID COPY
        server.EnqueueTaggedResponse("$tag NO failed\r\n");

        TestUtils.ExpectExceptionThrown<Smdn.Net.Imap4.Client.Session.ImapErrorResponseException>(delegate {
          // STORE won't be issued
          messages.MoveTo(destMailbox);
        });

        Assert.That(server.DequeueRequest(), Text.EndsWith("UID COPY 1,3,5 \"dest\"\r\n"));
      });
    }

    [Test, Ignore("not implemented")]
    public void TestCopyToDestinationMailboxUidValidityChanged()
    {
    }

    [Test, Ignore("not implemented")]
    public void TestMoveToDestinationMailboxUidValidityChanged()
    {
    }

    [Test, ExpectedException(typeof(ImapProtocolViolationException))]
    public void TestCopyToDestinationMailboxNonExistent()
    {
      CopyOrMoveToDestinationMailboxNonExistent(false);
    }

    [Test, ExpectedException(typeof(ImapProtocolViolationException))]
    public void TestMoveToDestinationMailboxNonExistent()
    {
      CopyOrMoveToDestinationMailboxNonExistent(true);
    }

    private void CopyOrMoveToDestinationMailboxNonExistent(bool move)
    {
      TestMessage(new[] {ImapCapability.ListExtended},
                  ImapSequenceSet.CreateUidRangeSet(1L, 10L),
                  delegate(ImapPseudoServer server, ImapMessageInfoBase messages) {
        // LIST
        server.EnqueueTaggedResponse("* LIST (\\Subscribed \\NonExistent) \"\" dest\r\n" +
                                     "$tag OK done\r\n");

        var destMailbox = messages.Client.GetMailbox("dest", ImapMailboxListOptions.SubscribedOnly);

        server.DequeueRequest();

        Assert.IsFalse(destMailbox.Exists);

        // throws exception
        if (move)
          messages.MoveTo(destMailbox);
        else
          messages.CopyTo(destMailbox);
      });
    }

    [Test]
    public void TestCopyToDestinationMailboxEmptySet()
    {
      CopyOrMoveToDestinationMailboxEmptySet(false);
    }

    [Test]
    public void TestMoveToDestinationMailboxEmptySet()
    {
      CopyOrMoveToDestinationMailboxEmptySet(true);
    }

    private void CopyOrMoveToDestinationMailboxEmptySet(bool move)
    {
      TestMessage(CreateEmptySet(),
                  delegate(ImapPseudoServer server, ImapMessageInfoBase messages) {
        // LIST
        server.EnqueueTaggedResponse("* LIST () \"\" dest\r\n" +
                                     "$tag OK done\r\n");

        var destMailbox = messages.Client.GetMailbox("dest");

        server.DequeueRequest();

        // do nothing
        if (move)
          messages.MoveTo(destMailbox);
        else
          messages.CopyTo(destMailbox);
      });
    }

    [Test, ExpectedException(typeof(ImapMailboxClosedException))]
    public void TestCopyToDestinationMailboxMailboxClosed()
    {
      CopyOrMoveToDestinationMailboxMailboxClosed(false);
    }

    [Test, ExpectedException(typeof(ImapMailboxClosedException))]
    public void TestMoveToDestinationMailboxMailboxClosed()
    {
      CopyOrMoveToDestinationMailboxMailboxClosed(true);
    }

    private void CopyOrMoveToDestinationMailboxMailboxClosed(bool move)
    {
      TestMessage(ImapSequenceSet.CreateUidSet(1L, 3L, 5L),
                  delegate(ImapPseudoServer server, ImapMessageInfoBase messages) {
        // LIST
        server.EnqueueTaggedResponse("* LIST () \"\" dest\r\n" +
                                     "$tag OK done\r\n");

        var destMailbox = messages.Client.GetMailbox("dest");

        server.DequeueRequest();

        // CLOSE
        server.EnqueueTaggedResponse("$tag OK done\r\n");

        messages.Mailbox.Close();

        Assert.IsFalse(messages.Mailbox.IsOpen);

        server.DequeueRequest();

        // throws exception
        if (move)
          messages.MoveTo(destMailbox);
        else
          messages.CopyTo(destMailbox);
      });
    }

    [Test]
    public void TestCopyToDestinationName()
    {
      CopyOrMoveToDestinationName(false, false);
    }

    [Test]
    public void TestMoveToDestinationName()
    {
      CopyOrMoveToDestinationName(true, false);
    }

    [Test]
    public void TestCopyToDestinationNameTryCreate()
    {
      CopyOrMoveToDestinationName(false, true);
    }

    [Test]
    public void TestMoveToDestinationNameTryCreate()
    {
      CopyOrMoveToDestinationName(true, true);
    }

    private void CopyOrMoveToDestinationName(bool move, bool tryCreate)
    {
      TestMessage(ImapSequenceSet.CreateUidRangeSet(1L, 10L),
                  delegate(ImapPseudoServer server, ImapMessageInfoBase messages) {
        // UID COPY
        server.EnqueueTaggedResponse("$tag OK done\r\n");
        if (move)
          // UID STORE
          server.EnqueueTaggedResponse("$tag OK done\r\n");

        ImapMailboxInfo createdMailbox = null;

        if (tryCreate) {
          createdMailbox = move
            ? messages.MoveTo("dest", true)
            : messages.CopyTo("dest", true);
        }
        else {
          if (move)
            messages.MoveTo("dest");
          else
            messages.CopyTo("dest");
        }
        // TODO: fetch updated attributes

        if (tryCreate)
          Assert.IsNull(createdMailbox);

        Assert.That(server.DequeueRequest(), Text.EndsWith("UID COPY 1:10 \"dest\"\r\n"));

        if (move)
          Assert.That(server.DequeueRequest(), Text.EndsWith("UID STORE 1:10 +FLAGS (\\Deleted)\r\n"));
      });
    }

    [Test]
    public void TestCopyToDestinationNameEmptySet()
    {
      CopyOrMoveToDestinationNameEmptySet(false);
    }

    [Test]
    public void TestMoveToDestinationNameEmptySet()
    {
      CopyOrMoveToDestinationNameEmptySet(true);
    }

    private void CopyOrMoveToDestinationNameEmptySet(bool move)
    {
      TestMessage(CreateEmptySet(),
                  delegate(ImapPseudoServer server, ImapMessageInfoBase messages) {
        // do nothing
        if (move)
          messages.MoveTo("dest");
        else
          messages.CopyTo("dest");
      });
    }

    [Test, Ignore("not implemented")]
    public void TestCopyToDestinationNameUidValidityChanged()
    {
    }

    [Test, Ignore("not implemented")]
    public void TestMoveToDestinationNameUidValidityChanged()
    {
    }

    [Test, ExpectedException(typeof(ImapMailboxClosedException))]
    public void TestCopyToDestinationNameMailboxClosed()
    {
      CopyOrMoveToDestinationNameMailboxClosed(false);
    }

    [Test, ExpectedException(typeof(ImapMailboxClosedException))]
    public void TestMoveToDestinationNameMailboxClosed()
    {
      CopyOrMoveToDestinationNameMailboxClosed(true);
    }

    private void CopyOrMoveToDestinationNameMailboxClosed(bool move)
    {
      TestMessage(ImapSequenceSet.CreateUidRangeSet(1L, 10L),
                  delegate(ImapPseudoServer server, ImapMessageInfoBase messages) {
        // CLOSE
        server.EnqueueTaggedResponse("$tag OK done\r\n");

        messages.Mailbox.Close();

        Assert.IsFalse(messages.Mailbox.IsOpen);

        server.DequeueRequest();

        // throws exception
        if (move)
          messages.MoveTo("dest");
        else
          messages.CopyTo("dest");
      });
    }

    [Test, ExpectedException(typeof(Smdn.Net.Imap4.Client.Session.ImapErrorResponseException))]
    public void TestCopyToDestinationNameNoResponse()
    {
      CopyOrMoveToDestinationNameNoResponse(false);
    }

    [Test, ExpectedException(typeof(Smdn.Net.Imap4.Client.Session.ImapErrorResponseException))]
    public void TestMoveToDestinationNameNoResponse()
    {
      CopyOrMoveToDestinationNameNoResponse(true);
    }

    private void CopyOrMoveToDestinationNameNoResponse(bool move)
    {
      TestMessage(ImapSequenceSet.CreateUidRangeSet(1L, 10L),
                  delegate(ImapPseudoServer server, ImapMessageInfoBase messages) {
        // UID COPY
        server.EnqueueTaggedResponse("$tag NO error occurred\r\n");

        // throws exception
        if (move)
          messages.MoveTo("dest", false);
        else
          messages.CopyTo("dest", false);
      });
    }

    [Test, ExpectedException(typeof(ImapMailboxNotFoundException))]
    public void TestCopyToDestinationNameNoResponseDontTryCreate()
    {
      CopyOrMoveToDestinationNameNoResponseDontTryCreate(false);
    }

    [Test, ExpectedException(typeof(ImapMailboxNotFoundException))]
    public void TestMoveToDestinationNameNoResponseDontTryCreate()
    {
      CopyOrMoveToDestinationNameNoResponseDontTryCreate(true);
    }

    private void CopyOrMoveToDestinationNameNoResponseDontTryCreate(bool move)
    {
      TestMessage(ImapSequenceSet.CreateUidRangeSet(1L, 10L),
                  delegate(ImapPseudoServer server, ImapMessageInfoBase messages) {
        // UID COPY
        server.EnqueueTaggedResponse("$tag NO [TRYCREATE] mailbox not exist\r\n");

        // throws exception
        if (move)
          messages.MoveTo("dest", false);
        else
          messages.CopyTo("dest", false);
      });
    }

    [Test]
    public void TestCopyToDestinationNameNoResponseTryCreate()
    {
      CopyOrMoveToDestinationNameNoResponseTryCreate(false);
    }

    [Test]
    public void TestMoveToDestinationNameNoResponseTryCreate()
    {
      CopyOrMoveToDestinationNameNoResponseTryCreate(true);
    }

    private void CopyOrMoveToDestinationNameNoResponseTryCreate(bool move)
    {
      TestMessage(ImapSequenceSet.CreateUidRangeSet(1L, 10L),
                  delegate(ImapPseudoServer server, ImapMessageInfoBase messages) {
        // UID COPY
        server.EnqueueTaggedResponse("$tag NO [TRYCREATE] mailbox not exist\r\n");
        // CREATE
        server.EnqueueTaggedResponse("$tag OK done\r\n");
        // UID COPY
        server.EnqueueTaggedResponse("$tag OK done\r\n");
        // LIST
        server.EnqueueTaggedResponse("* LIST () \"\" dest\r\n" +
                                     "$tag OK done\r\n");
        if (move)
          // UID STORE
          server.EnqueueTaggedResponse("$tag OK done\r\n");

        var createdMailbox = move
          ? messages.MoveTo("dest", true)
          : messages.CopyTo("dest", true);
        // TODO: fetch updated attributes

        Assert.IsNotNull(createdMailbox);
        Assert.AreEqual("dest", createdMailbox.Name);
        Assert.AreEqual("dest", createdMailbox.FullName);

        Assert.That(server.DequeueRequest(), Text.EndsWith("UID COPY 1:10 \"dest\"\r\n"));
        Assert.That(server.DequeueRequest(), Text.EndsWith("CREATE \"dest\"\r\n"));
        Assert.That(server.DequeueRequest(), Text.EndsWith("UID COPY 1:10 \"dest\"\r\n"));
        Assert.That(server.DequeueRequest(), Text.EndsWith("LIST \"\" \"dest\"\r\n"));

        if (move)
          Assert.That(server.DequeueRequest(), Text.EndsWith("UID STORE 1:10 +FLAGS (\\Deleted)\r\n"));
      });
    }
  }
}
