using System;
using System.Collections.Generic;
using NUnit.Framework;
using NUnit.Framework.SyntaxHelpers;

namespace Smdn.Net.Imap4.Client {
  [TestFixture]
  public class ImapOpenedMailboxInfoTests {
    private void TestOpenedMailbox(string mailboxName, Action<ImapPseudoServer, ImapOpenedMailboxInfo> action)
    {
      TestOpenedMailbox(null, mailboxName, "$tag OK [READ-WRITE] done\r\n", action);
    }

    private void TestOpenedMailbox(string mailboxName, string selectResponse, Action<ImapPseudoServer, ImapOpenedMailboxInfo> action)
    {
      TestOpenedMailbox(null, mailboxName, selectResponse, action);
    }

    private void TestOpenedMailbox(ImapCapability[] capabilities, string mailboxName, string selectResponse, Action<ImapPseudoServer, ImapOpenedMailboxInfo> action)
    {
      TestUtils.TestAuthenticated(capabilities, delegate(ImapPseudoServer server, ImapClient client) {
        // LIST
        server.EnqueueTaggedResponse("* LIST () \"\" INBOX\r\n" +
                                     "$tag OK done\r\n");
        // SELECT
        server.EnqueueTaggedResponse(selectResponse);

        using (var opened = client.OpenMailbox("INBOX")) {
          server.DequeueRequest(); // LIST
          server.DequeueRequest(); // SELECT

          try {
            Assert.IsTrue(opened.IsOpen);
            //Assert.AreEqual(opened.IsReadOnly, readOnly);
            Assert.IsNotNull(opened.ApplicableFlags);
            Assert.IsNotNull(opened.PermanentFlags);

            action(server, opened);
          }
          finally {
            // CLOSE
            if (opened.IsOpen)
              server.EnqueueTaggedResponse("$tag OK done\r\n");
          }
        }
      });
    }

    private void TestClosedMailbox(string mailboxName, Action<ImapPseudoServer, ImapOpenedMailboxInfo> action)
    {
      TestOpenedMailbox(mailboxName, delegate(ImapPseudoServer server, ImapOpenedMailboxInfo mailbox) {
        // CLOSE
        server.EnqueueTaggedResponse("$tag OK done\r\n");

        mailbox.Close();

        Assert.IsFalse(mailbox.IsOpen);
        Assert.IsNull(mailbox.Client.OpenedMailbox);

        Assert.That(server.DequeueRequest(), Text.EndsWith("CLOSE\r\n"));

        action(server, mailbox);
      });
    }

    [Test]
    public void TestRefresh()
    {
      TestOpenedMailbox("INBOX", delegate(ImapPseudoServer server, ImapOpenedMailboxInfo mailbox) {
        // NOOP
        server.EnqueueTaggedResponse("$tag OK done\r\n");

        mailbox.Refresh();
        // TODO: updated attributes, status, etc

        Assert.That(server.DequeueRequest(), Text.EndsWith("NOOP\r\n"));
      });
    }

    [Test, ExpectedException(typeof(ImapMailboxClosedException))]
    public void TestRefreshClosed()
    {
      TestClosedMailbox("INBOX", delegate(ImapPseudoServer server, ImapOpenedMailboxInfo mailbox) {
        mailbox.Refresh();
      });
    }

    [Test]
    public void TestOpen()
    {
      TestOpenedMailbox("INBOX", delegate(ImapPseudoServer server, ImapOpenedMailboxInfo mailbox) {
        Assert.IsTrue(mailbox.IsOpen);

        Assert.AreSame(mailbox, mailbox.Open());
        Assert.AreSame(mailbox, mailbox.Client.OpenedMailbox);

        Assert.IsTrue(mailbox.IsOpen);
      });
    }

    [Test]
    public void TestOpenAsReadOnly()
    {
      var selectResp =
        "$tag OK [READ-ONLY] done\r\n";

      TestOpenedMailbox("INBOX", selectResp, delegate(ImapPseudoServer server, ImapOpenedMailboxInfo mailbox) {
        Assert.IsTrue(mailbox.IsOpen);

        Assert.AreSame(mailbox, mailbox.Open());
        Assert.AreSame(mailbox, mailbox.Client.OpenedMailbox);
        Assert.IsTrue(mailbox.IsOpen);
        Assert.IsTrue(mailbox.IsReadOnly);
      });
    }

    [Test]
    public void TestOpenUidNotSticky()
    {
      var selectResp =
        "* OK [UIDVALIDITY 23]\r\n" +
        "* NO [UIDNOTSTICKY]\r\n" +
        "* FLAGS (\\Answered \\Flagged \\Deleted \\Seen \\Draft)\r\n" +
        "* OK [PERMANENTFLAGS (\\Deleted \\Seen)] Limited\r\n" +
        "$tag OK [READ-WRITE] done\r\n";

      TestOpenedMailbox("INBOX", selectResp, delegate(ImapPseudoServer server, ImapOpenedMailboxInfo mailbox) {
        Assert.IsTrue(mailbox.IsOpen);

        Assert.AreSame(mailbox, mailbox.Open());
        Assert.AreSame(mailbox, mailbox.Client.OpenedMailbox);
        Assert.IsTrue(mailbox.IsOpen);
        Assert.IsFalse(mailbox.IsReadOnly);
        Assert.AreEqual(23L, mailbox.UidValidity);
        Assert.IsFalse(mailbox.IsUidPersistent);

        Assert.IsNotNull(mailbox.ApplicableFlags);
        Assert.AreEqual(5, mailbox.ApplicableFlags.Count);

        Assert.IsNotNull(mailbox.PermanentFlags);
        Assert.AreEqual(2, mailbox.PermanentFlags.Count);
        Assert.IsFalse(mailbox.PermanentFlags.Has(ImapMessageFlag.AllowedCreateKeywords));
        Assert.IsFalse(mailbox.IsAllowedToCreateKeywords);
      });
    }

    [Test, Ignore("not implemented")]
    public void TestOpenNewly()
    {
    }

    [Test, Ignore("not implemented")]
    public void TestOpenReopenClosed()
    {
    }

    [Test, Ignore("not implemented")]
    public void TestOpenReopenReadWriteChange()
    {
    }

    [Test]
    public void TestClose()
    {
      TestOpenedMailbox("INBOX", delegate(ImapPseudoServer server, ImapOpenedMailboxInfo mailbox) {
        Assert.IsTrue(mailbox.IsOpen);

        // CLOSE
        server.EnqueueTaggedResponse("$tag OK done\r\n");

        mailbox.Close();
        // TODO: updated attributes, status, etc

        Assert.That(server.DequeueRequest(), Text.EndsWith("CLOSE\r\n"));

        Assert.IsFalse(mailbox.IsOpen);
        Assert.IsNull(mailbox.Client.OpenedMailbox);
      });
    }

    [Test]
    public void TestCloseAlreadyClosed()
    {
      TestClosedMailbox("INBOX", delegate(ImapPseudoServer server, ImapOpenedMailboxInfo mailbox) {
        // no exceptions
        mailbox.Close();

        Assert.IsFalse(mailbox.IsOpen);
        Assert.IsNull(mailbox.Client.OpenedMailbox);
      });
    }

    [Test]
    public void TestExpunge()
    {
      TestOpenedMailbox("INBOX", delegate(ImapPseudoServer server, ImapOpenedMailboxInfo mailbox) {
        // EXPUNGE
        server.EnqueueTaggedResponse("$tag OK done\r\n");

        mailbox.Expunge();
        // TODO: updated attributes, status, etc

        Assert.That(server.DequeueRequest(), Text.EndsWith("EXPUNGE\r\n"));
      });
    }

    [Test, ExpectedException(typeof(ImapMailboxClosedException))]
    public void TestExpungeClosed()
    {
      TestClosedMailbox("INBOX", delegate(ImapPseudoServer server, ImapOpenedMailboxInfo mailbox) {
        mailbox.Expunge();
      });
    }

    [Test]
    public void TestGetMessages()
    {
      var selectResp =
        "* 4 EXISTS\r\n" +
        "* OK [UIDVALIDITY 23]\r\n" +
        "$tag OK [READ-WRITE] done\r\n";

      TestOpenedMailbox("INBOX", selectResp, delegate(ImapPseudoServer server, ImapOpenedMailboxInfo mailbox) {
        var messages = mailbox.GetMessages();

        Assert.AreEqual(4L, mailbox.ExistMessageCount);

        using (var enumerator = messages.GetEnumerator()) {
          // NOOP
          server.EnqueueTaggedResponse("* 4 EXPUNGE\r\n" +
                                       "* 3 EXISTS\r\n" + 
                                       "$tag OK done\r\n");
          // FETCH
          server.EnqueueTaggedResponse("* FETCH 1 (UID 3)\r\n" +
                                       "* FETCH 2 (UID 5)\r\n" +
                                       "* FETCH 3 (UID 7)\r\n" +
                                       "$tag OK done\r\n");

          Assert.IsTrue(enumerator.MoveNext());

          Assert.That(server.DequeueRequest(), Text.EndsWith("NOOP\r\n"));
          Assert.That(server.DequeueRequest(), Text.EndsWith("FETCH 1,2,3 (UID)\r\n"));

          Assert.AreEqual(3L, mailbox.ExistMessageCount);

          Assert.IsNotNull(enumerator.Current, "#0 non-null");
          Assert.AreEqual(3L, enumerator.Current.Uid, "#0 uid");
          Assert.AreEqual(1L, enumerator.Current.Sequence, "#0 seq");

          Assert.IsTrue(enumerator.MoveNext());

          Assert.IsNotNull(enumerator.Current, "#1 non-null");
          Assert.AreEqual(5L, enumerator.Current.Uid, "#1 uid");
          Assert.AreEqual(2L, enumerator.Current.Sequence, "#1 seq");

          Assert.IsTrue(enumerator.MoveNext());

          Assert.IsNotNull(enumerator.Current, "#2 non-null");
          Assert.AreEqual(7L, enumerator.Current.Uid, "#2 uid");
          Assert.AreEqual(3L, enumerator.Current.Sequence, "#2 seq");

          Assert.IsFalse(enumerator.MoveNext());
        }

        /*
         * re-enumerate
         */
        Assert.AreEqual(3L, mailbox.ExistMessageCount);

        using (var enumerator = messages.GetEnumerator()) {
          // NOOP
          server.EnqueueTaggedResponse("* 2 EXPUNGE\r\n" +
                                       "* 1 EXPUNGE\r\n" +
                                       "$tag OK done\r\n");
          // FETCH
          server.EnqueueTaggedResponse("* FETCH 1 (UID 7)\r\n" +
                                       "$tag OK done\r\n");

          Assert.IsTrue(enumerator.MoveNext());

          Assert.That(server.DequeueRequest(), Text.EndsWith("NOOP\r\n"));
          Assert.That(server.DequeueRequest(), Text.EndsWith("FETCH 1 (UID)\r\n"));

          Assert.AreEqual(1L, mailbox.ExistMessageCount);

          Assert.IsNotNull(enumerator.Current, "#0 non-null");
          Assert.AreEqual(7L, enumerator.Current.Uid, "#0 uid");
          Assert.AreEqual(1L, enumerator.Current.Sequence, "#0 seq");

          Assert.IsFalse(enumerator.MoveNext());
        }
      });
    }

    [Test]
    public void TestGetMessagesNoMessageExists()
    {
      var selectResp =
        "* 2 EXISTS\r\n" +
        "* OK [UIDVALIDITY 23]\r\n" +
        "$tag OK [READ-WRITE] done\r\n";

      TestOpenedMailbox("INBOX", selectResp, delegate(ImapPseudoServer server, ImapOpenedMailboxInfo mailbox) {
        // NOOP
        server.EnqueueTaggedResponse("* 1 EXPUNGE\r\n" +
                                     "* 1 EXPUNGE\r\n" +
                                     "$tag OK done\r\n");

        var messages = new List<ImapMessageInfo>(mailbox.GetMessages());

        Assert.AreEqual(0L, mailbox.ExistMessageCount);

        Assert.AreEqual(0, messages.Count);

        Assert.That(server.DequeueRequest(), Text.EndsWith("NOOP\r\n"));
      });
    }

    [Test]
    public void TestGetMessageByUid()
    {
      GetMessageByUidOrSequence(true);
    }

    [Test]
    public void TestGetMessageBySequence()
    {
      GetMessageByUidOrSequence(false);
    }

    private void GetMessageByUidOrSequence(bool uid)
    {
      var selectResp =
        "* 1 EXISTS\r\n" +
        "* OK [UIDVALIDITY 23]\r\n" +
        "$tag OK [READ-WRITE] done\r\n";

      TestOpenedMailbox("INBOX", selectResp, delegate(ImapPseudoServer server, ImapOpenedMailboxInfo mailbox) {
        // FETCH
        server.EnqueueTaggedResponse("* FETCH 1 (UID 3)\r\n" +
                                     "$tag OK done\r\n");

        var message = uid
          ? mailbox.GetMessageByUid(3L)
          : mailbox.GetMessageBySequence(1L);

        Assert.IsNotNull(message);
        Assert.AreEqual(3L, message.Uid);
        Assert.AreEqual(1L, message.Sequence);
        Assert.AreEqual(23L, message.UidValidity);
        Assert.AreSame(mailbox, message.Mailbox);

        if (uid)
          Assert.That(server.DequeueRequest(), Text.EndsWith("UID FETCH 3 (UID)\r\n"));
        else
          Assert.That(server.DequeueRequest(), Text.EndsWith("FETCH 1 (UID)\r\n"));
      });
    }

    [Test, ExpectedException(typeof(ImapMessageNotFoundException))]
    public void TestGetMessageByUidNotFound()
    {
      var selectResp =
        "* 0 EXISTS\r\n" +
        "* OK [UIDVALIDITY 23]\r\n" +
        "$tag OK [READ-WRITE] done\r\n";

      TestOpenedMailbox("INBOX", selectResp, delegate(ImapPseudoServer server, ImapOpenedMailboxInfo mailbox) {
        // FETCH
        server.EnqueueTaggedResponse("$tag OK done\r\n");

        mailbox.GetMessageByUid(3L);
      });
    }

    [Test, ExpectedException(typeof(ArgumentOutOfRangeException))]
    public void TestGetMessageBySequenceGreaterThanExistCount()
    {
      var selectResp =
        "* 0 EXISTS\r\n" +
        "* OK [UIDVALIDITY 23]\r\n" +
        "$tag OK [READ-WRITE] done\r\n";

      TestOpenedMailbox("INBOX", selectResp, delegate(ImapPseudoServer server, ImapOpenedMailboxInfo mailbox) {
        mailbox.GetMessageBySequence(1L);
      });
    }

    [Test, ExpectedException(typeof(Smdn.Net.Imap4.Client.Session.ImapErrorResponseException))]
    public void TestGetMessageBySequenceNoResponse()
    {
      var selectResp =
        "* 1 EXISTS\r\n" +
        "* OK [UIDVALIDITY 23]\r\n" +
        "$tag OK [READ-WRITE] done\r\n";

      TestOpenedMailbox("INBOX", selectResp, delegate(ImapPseudoServer server, ImapOpenedMailboxInfo mailbox) {
        // FETCH
        server.EnqueueTaggedResponse("$tag NO such message\r\n");

        mailbox.GetMessageBySequence(1L);
      });
    }

    [Test]
    public void TestGetMessagesWithSearchCriteria()
    {
      var selectResp =
        "* OK [UIDVALIDITY 23]\r\n" +
        "$tag OK [READ-WRITE] done\r\n";

      TestOpenedMailbox("INBOX", selectResp, delegate(ImapPseudoServer server, ImapOpenedMailboxInfo mailbox) {
        var messages = mailbox.GetMessages(ImapSearchCriteria.Recent);

        using (var enumerator = messages.GetEnumerator()) {
          // SEARCH
          server.EnqueueTaggedResponse("* SEARCH 1 2 3\r\n" +
                                       "$tag OK done\r\n");
          // FETCH
          server.EnqueueTaggedResponse("* FETCH 1 (UID 3)\r\n" +
                                       "* FETCH 2 (UID 5)\r\n" +
                                       "* FETCH 3 (UID 7)\r\n" +
                                       "$tag OK done\r\n");

          Assert.IsTrue(enumerator.MoveNext());

          Assert.That(server.DequeueRequest(), Text.EndsWith("SEARCH RECENT\r\n"));
          Assert.That(server.DequeueRequest(), Text.EndsWith("FETCH 1,2,3 (UID)\r\n"));

          Assert.IsNotNull(enumerator.Current, "#0 non-null");
          Assert.AreEqual(3L, enumerator.Current.Uid, "#0 uid");
          Assert.AreEqual(1L, enumerator.Current.Sequence, "#0 seq");

          Assert.IsTrue(enumerator.MoveNext());

          Assert.IsNotNull(enumerator.Current, "#1 non-null");
          Assert.AreEqual(5L, enumerator.Current.Uid, "#1 uid");
          Assert.AreEqual(2L, enumerator.Current.Sequence, "#1 seq");

          Assert.IsTrue(enumerator.MoveNext());

          Assert.IsNotNull(enumerator.Current, "#2 non-null");
          Assert.AreEqual(7L, enumerator.Current.Uid, "#2 uid");
          Assert.AreEqual(3L, enumerator.Current.Sequence, "#2 seq");

          Assert.IsFalse(enumerator.MoveNext());
        }

        /*
         * re-enumerate
         */
        using (var enumerator = messages.GetEnumerator()) {
          // SEARCH
          server.EnqueueTaggedResponse("* SEARCH 4\r\n" +
                                       "$tag OK done\r\n");
          // FETCH
          server.EnqueueTaggedResponse("* FETCH 4 (UID 8)\r\n" +
                                       "$tag OK done\r\n");

          Assert.IsTrue(enumerator.MoveNext());

          Assert.That(server.DequeueRequest(), Text.EndsWith("SEARCH RECENT\r\n"));
          Assert.That(server.DequeueRequest(), Text.EndsWith("FETCH 4 (UID)\r\n"));

          Assert.IsNotNull(enumerator.Current, "#0 non-null");
          Assert.AreEqual(8L, enumerator.Current.Uid, "#0 uid");
          Assert.AreEqual(4L, enumerator.Current.Sequence, "#0 seq");

          Assert.IsFalse(enumerator.MoveNext());
        }
      });
    }

    [Test]
    public void TestGetMessagesWithSearchCriteriaSearchresCapable()
    {
      var selectResp =
        "* OK [UIDVALIDITY 23]\r\n" +
        "$tag OK [READ-WRITE] done\r\n";

      TestOpenedMailbox(new[] {ImapCapability.ESearch, ImapCapability.Searchres},
                        "INBOX",
                        selectResp,
                        delegate(ImapPseudoServer server, ImapOpenedMailboxInfo mailbox) {
        var messages = mailbox.GetMessages(ImapSearchCriteria.Recent);

        using (var enumerator = messages.GetEnumerator()) {
          // SEARCH
          server.EnqueueTaggedResponse("$tag OK done\r\n");
          // FETCH
          server.EnqueueTaggedResponse("* FETCH 1 (UID 3)\r\n" +
                                       "$tag OK done\r\n");

          Assert.IsTrue(enumerator.MoveNext());

          Assert.That(server.DequeueRequest(), Text.EndsWith("SEARCH RETURN (SAVE) RECENT\r\n"));
          Assert.That(server.DequeueRequest(), Text.EndsWith("FETCH $ (UID)\r\n"));

          Assert.IsNotNull(enumerator.Current, "#0 non-null");
          Assert.AreEqual(3L, enumerator.Current.Uid, "#0 uid");
          Assert.AreEqual(1L, enumerator.Current.Sequence, "#0 seq");

          Assert.IsFalse(enumerator.MoveNext());
        }
      });
    }

    [Test]
    public void TestGetMessagesWithSearchCriteriaNothingMatched()
    {
      var selectResp =
        "* OK [UIDVALIDITY 23]\r\n" +
        "$tag OK [READ-WRITE] done\r\n";

      TestOpenedMailbox("INBOX", selectResp, delegate(ImapPseudoServer server, ImapOpenedMailboxInfo mailbox) {
        var messages = mailbox.GetMessages(ImapSearchCriteria.Recent);

        using (var enumerator = messages.GetEnumerator()) {
          // SEARCH
          server.EnqueueTaggedResponse("* SEARCH\r\n" +
                                       "$tag OK done\r\n");

          Assert.IsFalse(enumerator.MoveNext());

          Assert.That(server.DequeueRequest(), Text.EndsWith("SEARCH RECENT\r\n"));
        }
      });
    }

    [Test]
    public void TestGetSortedMessages()
    {
      var selectResp =
        "* OK [UIDVALIDITY 23]\r\n" +
        "$tag OK [READ-WRITE] done\r\n";

      TestOpenedMailbox(new[] {ImapCapability.Sort},
                        "INBOX",
                        selectResp,
                        delegate(ImapPseudoServer server, ImapOpenedMailboxInfo mailbox) {
        var messages = mailbox.GetSortedMessages(ImapSortCriteria.Arrival, ImapSearchCriteria.Recent);

        using (var enumerator = messages.GetEnumerator()) {
          // SORT
          server.EnqueueTaggedResponse("* SORT 1 4 2\r\n" +
                                       "$tag OK done\r\n");
          // FETCH
          server.EnqueueTaggedResponse("* FETCH 1 (UID 2)\r\n" +
                                       "* FETCH 2 (UID 4)\r\n" +
                                       "* FETCH 4 (UID 8)\r\n" +
                                       "$tag OK done\r\n");

          Assert.IsTrue(enumerator.MoveNext());

          Assert.That(server.DequeueRequest(), Text.EndsWith("SORT (ARRIVAL) utf-8 RECENT\r\n"));
          Assert.That(server.DequeueRequest(), Text.EndsWith("FETCH 1,4,2 (UID)\r\n"));

          Assert.IsNotNull(enumerator.Current, "#0 non-null");
          Assert.AreEqual(2L, enumerator.Current.Uid, "#0 uid");
          Assert.AreEqual(1L, enumerator.Current.Sequence, "#0 seq");

          Assert.IsTrue(enumerator.MoveNext());

          Assert.IsNotNull(enumerator.Current, "#1 non-null");
          Assert.AreEqual(4L, enumerator.Current.Uid, "#1 uid");
          Assert.AreEqual(2L, enumerator.Current.Sequence, "#1 seq");

          Assert.IsTrue(enumerator.MoveNext());

          Assert.IsNotNull(enumerator.Current, "#2 non-null");
          Assert.AreEqual(8L, enumerator.Current.Uid, "#2 uid");
          Assert.AreEqual(4L, enumerator.Current.Sequence, "#2 seq");

          Assert.IsFalse(enumerator.MoveNext());
        }
      });
    }

    [Test, ExpectedException(typeof(ImapIncapableException))]
    public void TestGetSortedMessagesSortIncapable()
    {
      var selectResp =
        "* OK [UIDVALIDITY 23]\r\n" +
        "$tag OK [READ-WRITE] done\r\n";

      TestOpenedMailbox("INBOX", selectResp, delegate(ImapPseudoServer server, ImapOpenedMailboxInfo mailbox) {
        mailbox.GetSortedMessages(ImapSortCriteria.Subject, ImapSearchCriteria.Recent);
      });
    }


    [Test]
    public void TestGetSortedMessagesNothingMatched()
    {
      var selectResp =
        "* OK [UIDVALIDITY 23]\r\n" +
        "$tag OK [READ-WRITE] done\r\n";

      TestOpenedMailbox(new[] {ImapCapability.Sort},
                        "INBOX",
                        selectResp,
                        delegate(ImapPseudoServer server, ImapOpenedMailboxInfo mailbox) {
        var messages = mailbox.GetSortedMessages(ImapSortCriteria.Arrival, ImapSearchCriteria.Recent);

        using (var enumerator = messages.GetEnumerator()) {
          // SORT
          server.EnqueueTaggedResponse("* SORT\r\n" +
                                       "$tag OK done\r\n");

          Assert.IsFalse(enumerator.MoveNext());

          Assert.That(server.DequeueRequest(), Text.EndsWith("SORT (ARRIVAL) utf-8 RECENT\r\n"));
        }
      });
    }

    [Test]
    public void TestWaitForMessageArrival()
    {
    }

    [Test]
    public void TestBeginWaitForMessageArrival()
    {
    }

    [Test]
    public void TestEndWaitForMessageArrival()
    {
    }
  }
}
